@@ -1,10 +1,15 @@
-use crate::{ModularCoreOps, ModularOps};
+use crate::{ModularCoreOps, ModularPow, ModularSymbols, ModularUnaryOps};
 use core::convert::TryInto;
 use num_integer::Integer;
 use num_traits::{One, ToPrimitive, Zero};
 
-macro_rules! impl_mod_arithm_by_ref {
+// Efficient implementation for bigints can be found in "Handbook of Applied Cryptography"
+// Reference: https://cacr.uwaterloo.ca/hac/about/chap14.pdf
+
+// Forward modular operations to ref by ref
+macro_rules! impl_mod_ops_by_ref {
     ($T:ty) => {
+        // core ops
         impl ModularCoreOps<$T, &$T> for &$T {
             type Output = $T;
             #[inline]
@@ -19,34 +24,7 @@ macro_rules! impl_mod_arithm_by_ref {
             fn mulm(self, rhs: $T, m: &$T) -> $T {
                 self.mulm(&rhs, &m)
             }
-            #[inline]
-            fn negm(self, m: &$T) -> $T {
-                ModularCoreOps::<&$T, &$T>::negm(self, m)
-            }
         }
-        impl ModularOps<$T, &$T> for &$T {
-            #[inline]
-            fn powm(self, exp: $T, m: &$T) -> $T {
-                self.powm(&exp, &m)
-            }
-            #[inline]
-            fn invm(self, m: &$T) -> Option<$T> {
-                ModularOps::<&$T, &$T>::invm(self, m)
-            }
-            #[inline]
-            fn legendre(self, n: &$T) -> i8 {
-                ModularOps::<&$T, &$T>::legendre(self, n)
-            }
-            #[inline]
-            fn jacobi(self, n: &$T) -> i8 {
-                ModularOps::<&$T, &$T>::jacobi(self, n)
-            }
-            #[inline]
-            fn kronecker(self, n: &$T) -> i8 {
-                ModularOps::<&$T, &$T>::kronecker(self, n)
-            }
-        }
-
         impl ModularCoreOps<&$T, &$T> for $T {
             type Output = $T;
             #[inline]
@@ -61,34 +39,7 @@ macro_rules! impl_mod_arithm_by_ref {
             fn mulm(self, rhs: &$T, m: &$T) -> $T {
                 (&self).mulm(rhs, &m)
             }
-            #[inline]
-            fn negm(self, m: &$T) -> $T {
-                ModularCoreOps::<&$T, &$T>::negm(&self, m)
-            }
         }
-        impl ModularOps<&$T, &$T> for $T {
-            #[inline]
-            fn powm(self, exp: &$T, m: &$T) -> $T {
-                (&self).powm(exp, &m)
-            }
-            #[inline]
-            fn invm(self, m: &$T) -> Option<$T> {
-                ModularOps::<&$T, &$T>::invm(&self, m)
-            }
-            #[inline]
-            fn legendre(self, n: &$T) -> i8 {
-                ModularOps::<&$T, &$T>::legendre(&self, n)
-            }
-            #[inline]
-            fn jacobi(self, n: &$T) -> i8 {
-                ModularOps::<&$T, &$T>::jacobi(&self, n)
-            }
-            #[inline]
-            fn kronecker(self, n: &$T) -> i8 {
-                ModularOps::<&$T, &$T>::kronecker(&self, n)
-            }
-        }
-
         impl ModularCoreOps<$T, &$T> for $T {
             type Output = $T;
             #[inline]
@@ -103,38 +54,56 @@ macro_rules! impl_mod_arithm_by_ref {
             fn mulm(self, rhs: $T, m: &$T) -> $T {
                 (&self).mulm(&rhs, &m)
             }
+        }
+
+        // pow
+        impl ModularPow<$T, &$T> for &$T {
+            type Output = $T;
             #[inline]
-            fn negm(self, m: &$T) -> $T {
-                ModularCoreOps::<&$T, &$T>::negm(&self, m)
+            fn powm(self, exp: $T, m: &$T) -> $T {
+                self.powm(&exp, &m)
             }
         }
-        impl ModularOps<$T, &$T> for $T {
+        impl ModularPow<&$T, &$T> for $T {
+            type Output = $T;
+            #[inline]
+            fn powm(self, exp: &$T, m: &$T) -> $T {
+                (&self).powm(exp, &m)
+            }
+        }
+        impl ModularPow<$T, &$T> for $T {
+            type Output = $T;
             #[inline]
             fn powm(self, exp: $T, m: &$T) -> $T {
                 (&self).powm(&exp, &m)
             }
+        }
+
+        // unary ops and symbols
+        impl ModularUnaryOps<&$T> for $T {
+            type Output = $T;
+            #[inline]
+            fn negm(self, m: &$T) -> $T {
+                ModularUnaryOps::<&$T>::negm(&self, m)
+            }
             #[inline]
             fn invm(self, m: &$T) -> Option<$T> {
-                ModularOps::<&$T, &$T>::invm(&self, m)
+                ModularUnaryOps::<&$T>::invm(&self, m)
             }
             #[inline]
-            fn legendre(self, n: &$T) -> i8 {
-                ModularOps::<&$T, &$T>::legendre(&self, n)
+            fn dblm(self, m: &$T) -> $T {
+                ModularUnaryOps::<&$T>::dblm(&self, m)
             }
             #[inline]
-            fn jacobi(self, n: &$T) -> i8 {
-                ModularOps::<&$T, &$T>::jacobi(&self, n)
-            }
-            #[inline]
-            fn kronecker(self, n: &$T) -> i8 {
-                ModularOps::<&$T, &$T>::kronecker(&self, n)
+            fn sqm(self, m: &$T) -> $T {
+                ModularUnaryOps::<&$T>::sqm(&self, m)
             }
         }
     };
 }
 
 #[cfg(feature = "num-bigint")]
-mod impl_num_bigint {
+mod _num_bigint {
     use super::*;
     use num_bigint::BigUint;
 
@@ -166,7 +135,10 @@ mod impl_num_bigint {
 
             (a * b) % m
         }
+    }
 
+    impl ModularUnaryOps<&BigUint> for &BigUint {
+        type Output = BigUint;
         #[inline]
         fn negm(self, m: &BigUint) -> BigUint {
             let x = self % m;
@@ -176,13 +148,6 @@ mod impl_num_bigint {
                 m - x
             }
         }
-    }
-
-    impl ModularOps<&BigUint, &BigUint> for &BigUint {
-        #[inline]
-        fn powm(self, exp: &BigUint, m: &BigUint) -> BigUint {
-            self.modpow(&exp, m)
-        }
 
         fn invm(self, m: &BigUint) -> Option<Self::Output> {
             let x = if self >= m { self % m } else { self.clone() };
@@ -209,28 +174,54 @@ mod impl_num_bigint {
         }
 
         #[inline]
-        fn legendre(self, n: &BigUint) -> i8 {
-            let r = self.powm((n - 1u8) >> 1u8, &n);
-            if r.is_zero() {
-                return 0;
+        fn dblm(self, m: &BigUint) -> BigUint {
+            let x = self % m;
+            let d = x << 1;
+            if &d > m {
+                d - m
+            } else {
+                d
             }
-            if r.is_one() {
-                return 1;
-            }
-            if &(r + 1u8) == n {
-                return -1;
-            }
-            panic!("n is not prime!")
         }
 
-        fn jacobi(self, n: &BigUint) -> i8 {
-            debug_assert!(n.is_odd());
+        #[inline]
+        fn sqm(self, m: &BigUint) -> BigUint {
+            self.mulm(self, m)
+        }
+    }
 
+    impl ModularPow<&BigUint, &BigUint> for &BigUint {
+        type Output = BigUint;
+        #[inline]
+        fn powm(self, exp: &BigUint, m: &BigUint) -> BigUint {
+            self.modpow(&exp, m)
+        }
+    }
+
+    impl ModularSymbols<&BigUint> for BigUint {
+        #[inline]
+        fn checked_legendre(&self, n: &BigUint) -> Option<i8> {
+            let r = self.powm((n - 1u8) >> 1u8, &n);
+            if r.is_zero() {
+                Some(0)
+            } else if r.is_one() {
+                Some(1)
+            } else if &(r + 1u8) == n {
+                Some(-1)
+            } else {
+                None
+            }
+        }
+
+        fn checked_jacobi(&self, n: &BigUint) -> Option<i8> {
+            if n.is_even() {
+                return None;
+            }
             if self.is_zero() {
-                return 0;
+                return Some(0);
             }
             if self.is_one() {
-                return 1;
+                return Some(1);
             }
 
             let three = BigUint::from(3u8);
@@ -253,15 +244,11 @@ mod impl_num_bigint {
                 }
                 a %= &n;
             }
-            if n.is_one() {
-                t
-            } else {
-                0
-            }
+            Some(if n.is_one() { t } else { 0 })
         }
 
         #[inline]
-        fn kronecker(self, n: &BigUint) -> i8 {
+        fn kronecker(&self, n: &BigUint) -> i8 {
             if n.is_zero() {
                 return if self.is_one() { 1 } else { 0 };
             }
@@ -283,11 +270,65 @@ mod impl_num_bigint {
 
             let f = n.trailing_zeros().unwrap_or(0);
             let n = n >> f;
-            let t1 = ModularOps::<&BigUint, &BigUint>::kronecker(self, &BigUint::from(2u8));
-            let t2 = ModularOps::<&BigUint, &BigUint>::jacobi(self, &n);
+            let t1 = self.kronecker(&BigUint::from(2u8));
+            let t2 = self.jacobi(&n);
             t1.pow(f.try_into().unwrap()) * t2
         }
     }
 
-    impl_mod_arithm_by_ref!(BigUint);
+    impl_mod_ops_by_ref!(BigUint);
+
+    #[cfg(test)]
+    mod tests {
+        use super::*;
+        use rand::random;
+
+        const NRANDOM: u32 = 10; // number of random tests to run
+
+        #[test]
+        fn basic_tests() {
+            for _ in 0..NRANDOM {
+                let a = random::<u128>();
+                let ra = &BigUint::from(a);
+                let b = random::<u128>();
+                let rb = &BigUint::from(b);
+                let m = random::<u128>() | 1;
+                let rm = &BigUint::from(m);
+                assert_eq!(ra.addm(rb, rm), (ra + rb) % rm);
+                assert_eq!(ra.mulm(rb, rm), (ra * rb) % rm);
+
+                let a = random::<u8>();
+                let ra = &BigUint::from(a);
+                let e = random::<u8>();
+                let re = &BigUint::from(e);
+                let m = random::<u128>() | 1;
+                let rm = &BigUint::from(m);
+                assert_eq!(ra.powm(re, rm), ra.pow(e as u32) % rm);
+            }
+        }
+
+        #[test]
+        fn test_against_prim() {
+            for _ in 0..NRANDOM {
+                let a = random::<u128>();
+                let ra = &BigUint::from(a);
+                let b = random::<u128>();
+                let rb = &BigUint::from(b);
+                let m = random::<u128>();
+                let rm = &BigUint::from(m);
+                assert_eq!(ra.addm(rb, rm), a.addm(b, &m).into());
+                assert_eq!(ra.subm(rb, rm), a.subm(b, &m).into());
+                assert_eq!(ra.mulm(rb, rm), a.mulm(b, &m).into());
+                assert_eq!(ra.negm(rm), a.negm(&m).into());
+                assert_eq!(ra.invm(rm), a.invm(&m).map(|v| v.into()));
+                assert_eq!(ra.checked_legendre(rm), a.checked_legendre(&m));
+                assert_eq!(ra.checked_jacobi(rm), a.checked_jacobi(&m));
+                assert_eq!(ra.kronecker(rm), a.kronecker(&m));
+
+                let e = random::<u8>();
+                let re = &BigUint::from(e);
+                assert_eq!(ra.powm(re, rm), a.powm(e as u128, &m).into());
+            }
+        }
+    }
 }
